from contextvars import ContextVar, Token
from typing import Optional, Type

from boto3.dynamodb.conditions import ConditionExpressionBuilder
from boto3.dynamodb.types import TypeSerializer

from . import main

# DynamoDB transaction commit limit
# https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/transaction-apis.html
TRANSACTION_MAX_ITEMS = 100

_transaction_writer_var: ContextVar[Optional["TransactionWriter"]] = ContextVar(
    "transaction_writer",
    default=None,
)


def current_transaction_writer() -> Optional["TransactionWriter"]:
    return _transaction_writer_var.get()


_dynamodb_serializer = TypeSerializer()
_dynamodb_builder = ConditionExpressionBuilder()


def serialize_data(item: dict) -> dict:
    return {k: _dynamodb_serializer.serialize(v) for k, v in item.items()}


def serialize_condition(condition) -> dict:
    expression = _dynamodb_builder.build_expression(condition)

    return {
        "ConditionExpression": expression.condition_expression,
        "ExpressionAttributeNames": expression.attribute_name_placeholders,
        "ExpressionAttributeValues": serialize_data(expression.attribute_value_placeholders),
    }


class TransactionWriter:
    def __init__(self, auto_commit: bool = False, commit_every: int = TRANSACTION_MAX_ITEMS):
        self.items: list = []
        self.auto_commit = auto_commit
        assert (
            commit_every <= TRANSACTION_MAX_ITEMS
        ), f"commit_every cannot exceed DynamoDB limit of {TRANSACTION_MAX_ITEMS} items"
        self.commit_max = commit_every
        self.batches_submitted = 0
        self._first_table: Optional[Type["main.Dyntastic"]] = None
        self._context_var_reset_token: Optional[Token] = None

    def __enter__(self):
        self._context_var_reset_token = _transaction_writer_var.set(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        assert self._context_var_reset_token is not None
        _transaction_writer_var.reset(self._context_var_reset_token)
        self._context_var_reset_token = None

        if not exc_type:
            self.commit()

    def commit(self):
        if self.items:
            assert self._first_table is not None, "If any items were added, there should be a table"
            client = self._first_table._dynamodb_client()
            client.transact_write_items(TransactItems=self.items)
            self.items = []
            self.batches_submitted += 1

    def _register_table(self, table: Type["main.Dyntastic"]):
        if self._first_table is None:
            self._first_table = table
        else:
            assert self._first_table.__table_region__ == table.__table_region__, (
                "All tables in a transaction must be in the same region "
                f"(found {self._first_table.__table_region__} and {table.__table_region__})"
            )

    def _at_max_items(self):
        return len(self.items) == TRANSACTION_MAX_ITEMS

    def _at_commit_max(self):
        return len(self.items) == self.commit_max

    def add(self, table: Type["main.Dyntastic"], item: dict):
        if not self.auto_commit and self._at_max_items():
            raise Exception(f"Exceeded max items ({TRANSACTION_MAX_ITEMS}) in transaction")

        self._register_table(table)
        self.items.append(item)

        if self.auto_commit and self._at_commit_max():
            self.commit()



from collections import defaultdict
from decimal import Decimal
from typing import Optional, Union

from boto3.dynamodb.conditions import Attr as _DynamoAttr
from boto3.dynamodb.conditions import Key as _DynamoKey
from boto3.dynamodb.types import TypeDeserializer
from pydantic import BaseModel

from . import pydantic_compat

# serialization helpers


# Except for sets and Decimal, pydantic_core.as_jsonable_python would work.
# To properly support these cases, however, we need to walk through the data.
def serialize(data):
    if isinstance(data, BaseModel):
        return serialize(pydantic_compat.model_dump(data))
    elif isinstance(data, dict):
        # TODO: May not actually want to filter out None. Without the filter,
        # all None fields in the pydantic model appear as Null instead of
        # nonexistent
        return {key: serialize(value) for key, value in data.items() if value is not None}
    elif isinstance(data, (list, tuple)):
        return list(map(serialize, data))
    elif isinstance(data, set):
        return set(map(serialize, data))
    elif isinstance(data, (Decimal, str, int, bytes, bool, float, type(None))):
        return data
    else:
        # handle types like datetime
        return pydantic_compat.to_jsonable_python(data)


# Hacky way to avoid boto3's annoying Binary type
# used as a wrapper around standard bytes objects
TypeDeserializer._deserialize_b = lambda _, value: value  # type: ignore


# condition helpers


def _ensure_value(value):
    if isinstance(value, _BaseAttr):
        return value._attr
    else:
        return serialize(value)


def _attr_method(method_name: str):
    def _attr_operation(self, value):
        method = getattr(self._attr, method_name)
        return method(_ensure_value(value))

    return _attr_operation


def _key_method(method_name: str):
    def _key_operation(self, value):
        method = getattr(self._key, method_name)
        return method(_ensure_value(value))

    return _key_operation


# update helpers


class _Variables:
    def __init__(self):
        self.attributes = {}
        self.values = {}

    def add_variable(self, variable):
        if isinstance(variable, _UpdateFn):
            return variable.build(self)
        elif isinstance(variable, Attr):
            # TODO: support indexes in the path as well (e.g. "my_list[0].nested_attr")
            data = self.attributes
            key_prefix = "#"
            # if the attribute is a nested path, DynamoDB expects each segment to be
            # in a separate entry in ExpressionAttributeNames
            variables = variable.name.split(".")
        else:
            data = self.values
            key_prefix = ":"
            variables = [variable]

        segments = []
        for var in variables:
            key = f"{key_prefix}{len(data)}"
            data[key] = var
            segments.append(key)

        return ".".join(segments)


class _UpdateAction:
    update_action: str

    def build(self, variables: _Variables) -> str:  # pragma: no cover
        raise NotImplementedError

    def __str__(self):
        variables = _Variables()
        serialized = self.build(variables)

        action = f"{self.update_action} " if hasattr(self, "update_action") else ""
        expression = f"Expression='{action}{serialized}'"
        attributes = f"Attributes={variables.attributes}"
        values = f"Values={variables.values}"

        return f"{self.__class__.__name__}({expression}, {attributes}, {values})"

    def __repr__(self):
        return str(self)


class _UpdatePathValue(_UpdateAction):
    def __init__(self, path, value):
        self.path = path
        self.value = value

    def build(self, variables: _Variables):
        path_var = variables.add_variable(self.path)
        value_var = variables.add_variable(self.value)
        return self.format(path_var, value_var)

    def format(self, path_var: str, value_var: str):
        return f"{path_var} {value_var}"


class _ActionSet(_UpdatePathValue):
    update_action = "SET"

    def format(self, path_var: str, value_var: str):
        return f"{path_var} = {value_var}"


class _UpdateFn(_UpdateAction):
    fn_name: str

    def __init__(self, *args):
        self.args = args

    def build(self, variables: _Variables):
        arg_vars = map(variables.add_variable, self.args)
        formatted_args = ", ".join(arg_vars)
        return f"{self.fn_name}({formatted_args})"


class _IfNotExists(_UpdateFn):
    fn_name = "if_not_exists"


class _ListAppend(_UpdateFn):
    fn_name = "list_append"


class _Operator(_UpdateFn):
    operator: str

    def __init__(self, arg1, arg2):
        self.arg1 = arg1
        self.arg2 = arg2

    def build(self, variables: _Variables):
        arg1_var = variables.add_variable(self.arg1)
        arg2_var = variables.add_variable(self.arg2)
        return f"{arg1_var} {self.operator} {arg2_var}"


class _Plus(_Operator):
    operator = "+"


class _Minus(_Operator):
    operator = "-"


class _ActionRemove(_UpdateAction):
    update_action = "REMOVE"

    def __init__(self, path, index: Optional[int] = None):
        # This uses `type(...) is not ...` instead of `isinstance` to
        # ensure that subtypes of int which might have a different
        # __str__ method cannot allow injection
        if index is not None and type(index) is not int:
            raise ValueError(f"Dyntastic remove() update must be given an int, found '{index.__class__.__name__}'")

        self.path = path
        self.index = index

    def build(self, variables: _Variables):
        path_var = variables.add_variable(self.path)

        if self.index is None:
            return path_var
        else:
            return f"{path_var}[{self.index}]"


class _ActionAdd(_UpdatePathValue):
    update_action = "ADD"


class _ActionDelete(_UpdatePathValue):
    update_action = "DELETE"


def translate_updates(*actions: _UpdateAction):
    variables = _Variables()
    serialized_actions = defaultdict(list)

    for action in actions:
        serialized_actions[action.update_action].append(action.build(variables))

    action_expressions = []
    for update_action, expressions in serialized_actions.items():
        joined_updates = ", ".join(expressions)
        action_expressions.append(f"{update_action} {joined_updates}")

    update_expression = " ".join(action_expressions)
    update_data = {
        "UpdateExpression": update_expression,
        "ExpressionAttributeNames": variables.attributes,
    }

    if variables.values:
        update_data["ExpressionAttributeValues"] = variables.values

    return update_data


# tying everything together


class _AttrMetaclass(type):
    def __getattr__(cls, name: str) -> "Attr":
        return cls(name)


class _BaseAttr(metaclass=_AttrMetaclass):
    def __init__(self, name: str):
        self.name = name
        self._key = _DynamoKey(name)
        self._attr = _DynamoAttr(name)

    def __str__(self):
        return f"Attr<{self.name}>"

    def __repr__(self):
        return str(self)

    # conditions that can work on a DynamoDB Key

    eq = __eq__ = _key_method("eq")
    lt = __lt__ = _key_method("lt")
    le = __le__ = _key_method("lte")
    gt = __gt__ = _key_method("gt")
    ge = __ge__ = _key_method("gte")
    begins_with = _key_method("begins_with")

    def between(self, low_value, high_value):
        return self._key.between(_ensure_value(low_value), _ensure_value(high_value))


class Attr(_BaseAttr):
    # conditions that can only work on a non-key attribute

    ne = __ne__ = _attr_method("ne")
    is_in = _attr_method("is_in")
    contains = _attr_method("contains")
    attribute_type = _attr_method("attribute_type")

    @property
    def size(self):
        return _Size(self.name)

    def exists(self):
        return self._attr.exists()

    def not_exists(self):
        return self._attr.not_exists()

    # update "subexpressions"

    def if_not_exists(self, value):
        return _IfNotExists(self, value)

    # TODO: allow _ListAppend to be created with list value in either position?
    def list_append(self: Union["Attr", list, tuple], value):
        if isinstance(value, tuple):
            value = list(value)
        elif not isinstance(value, list):
            value = [value]

        return _ListAppend(self, value)

    def __add__(self, value):
        return _Plus(self, value)

    def __radd__(self, value):
        return _Plus(value, self)

    plus = __add__

    def __sub__(self, value):
        return _Minus(self, value)

    def __rsub__(self, value):
        return _Minus(value, self)

    minus = __sub__

    # update expressions

    def set(self, value):
        return _ActionSet(self, value)

    def set_default(self, value):
        return self.set(self.if_not_exists(value))

    def append(self, value):
        return self.set(self.list_append(value))

    def remove(self, index: Optional[int] = None):
        return _ActionRemove(self, index)

    def add(self, value):
        if isinstance(value, (str, bytes)):
            value = {value}
        elif isinstance(value, (tuple, list)):
            value = set(value)
        return _ActionAdd(self, value)

    def delete(self, value):
        if isinstance(value, (str, bytes)):
            value = {value}
        elif isinstance(value, (tuple, list)):
            value = set(value)
        return _ActionDelete(self, value)


class _Size(_BaseAttr):
    def __init__(self, name: str):
        self.name = name
        self._key = _DynamoAttr(name).size()  # type: ignore
        self._attr = _DynamoAttr(name).size()  # type: ignore


# Alias for ease of use
A = Attr



from decimal import Decimal

import pytest

from dyntastic import A
from dyntastic.attr import translate_updates


def _assert_expression(expression, attributes, values, *update_actions):
    translated_attributes = {f"#{i}": attr for i, attr in enumerate(attributes)}
    translated_values = {f":{i}": value for i, value in enumerate(values)}

    expected = {"UpdateExpression": expression}
    if translated_attributes:
        expected["ExpressionAttributeNames"] = translated_attributes
    if translated_values:
        expected["ExpressionAttributeValues"] = translated_values

    assert translate_updates(*update_actions) == expected


def test_update_action_str():
    assert (
        str(A.my_field.set(5))
        == repr(A.my_field.set(5))
        == "_ActionSet(Expression='SET #0 = :0', Attributes={'#0': 'my_field'}, Values={':0': 5})"
    )


def test_attr_str():
    assert str(A.my_field) == repr(A.my_field) == "Attr<my_field>"
    assert str(A("my_field.nested_field")) == repr(A("my_field.nested_field")) == "Attr<my_field.nested_field>"
    assert (
        str(A("my_field.nested_field.nested2"))
        == repr(A("my_field.nested_field.nested2"))
        == "Attr<my_field.nested_field.nested2>"
    )


def test_set_value():
    _assert_expression("SET #0 = :0", ["my_str"], ["my_value"], A.my_str.set("my_value"))
    _assert_expression("SET #0.#1 = :0", ["my_dict", "my_str"], ["my_value"], A("my_dict.my_str").set("my_value"))
    _assert_expression(
        "SET #0.#1.#2 = :0",
        ["my_dict", "nested_dict", "my_str"],
        ["my_value"],
        A("my_dict.nested_dict.my_str").set("my_value"),
    )


def test_set_attribute():
    _assert_expression("SET #0 = #1", ["my_str", "my_other_str"], [], A.my_str.set(A.my_other_str))
    _assert_expression(
        "SET #0.#1 = #2", ["my_dict", "my_str", "my_other_str"], [], A("my_dict.my_str").set(A.my_other_str)
    )
    _assert_expression(
        "SET #0.#1.#2 = #3.#4.#5",
        ["my_dict1", "nested1", "my_str", "my_dict2", "nested2", "my_str2"],
        [],
        A("my_dict1.nested1.my_str").set(A("my_dict2.nested2.my_str2")),
    )


def test_set_multiple_attributes_and_values():
    _assert_expression(
        "SET #0 = #1, #2 = :0, #3.#4 = #5.#6.#7",
        ["my_str", "my_other_str", "my_int", "my_dict", "nested_str", "dict2", "nested_dict", "nested_str2"],
        [1],
        A.my_str.set(A.my_other_str),
        A.my_int.set(1),
        A("my_dict.nested_str").set(A("dict2.nested_dict.nested_str2")),
    )


def test_set_default():
    expected = ("SET #0 = if_not_exists(#1, :0)", ["my_int", "my_int"], [100])
    _assert_expression(*expected, A.my_int.set_default(100))
    _assert_expression(*expected, A.my_int.set(A.my_int.if_not_exists(100)))

    expected = ("SET #0 = if_not_exists(#1, #2)", ["my_int", "my_int", "my_other_int"], [])
    _assert_expression(*expected, A.my_int.set_default(A.my_other_int))
    _assert_expression(*expected, A.my_int.set(A.my_int.if_not_exists(A.my_other_int)))

    expected = ("SET #0.#1 = if_not_exists(#2.#3, #4)", ["my_dict", "my_int", "my_dict", "my_int", "my_other_int"], [])
    _assert_expression(*expected, A("my_dict.my_int").set_default(A.my_other_int))
    _assert_expression(*expected, A("my_dict.my_int").set(A("my_dict.my_int").if_not_exists(A.my_other_int)))


def test_if_not_exists():
    _assert_expression(
        "SET #0 = if_not_exists(#1, :0)",
        ["my_field", "my_other_field"],
        ["my_value"],
        A.my_field.set(A.my_other_field.if_not_exists("my_value")),
    )

    _assert_expression(
        "SET #0.#1 = if_not_exists(#2.#3, :0)",
        ["my_dict", "my_field", "my_other_dict", "my_other_field"],
        ["my_value"],
        A("my_dict.my_field").set(A("my_other_dict.my_other_field").if_not_exists("my_value")),
    )


def test_append():
    expected = ("SET #0 = list_append(#1, :0)", ["my_field", "my_field"], [[1, 2]])
    _assert_expression(*expected, A.my_field.append([1, 2]))
    _assert_expression(*expected, A.my_field.append((1, 2)))


def test_append_single_value():
    expected = ("SET #0 = list_append(#1, :0)", ["my_field", "my_field"], [[1]])
    _assert_expression(*expected, A.my_field.append(1))

    expected = ("SET #0.#1 = list_append(#2.#3, :0)", ["my_dict", "my_field", "my_dict", "my_field"], [[1]])
    _assert_expression(*expected, A("my_dict.my_field").append(1))


def test_list_append_two_values():
    _assert_expression(
        "SET #0 = list_append(:0, :1)",
        ["my_field"],
        [[1, 2], [3, 4]],
        A.my_field.set(A.list_append([1, 2], [3, 4])),
    )

    _assert_expression(
        "SET #0.#1 = list_append(:0, :1)",
        ["my_dict", "my_field"],
        [[1, 2], [3, 4]],
        A("my_dict.my_field").set(A.list_append([1, 2], [3, 4])),
    )


def test_set_plus_value():
    _assert_expression("SET #0 = #1 + :0", ["my_field", "my_field"], [5], A.my_field.set(A.my_field + 5))
    _assert_expression("SET #0 = :0 + #1", ["my_field", "my_field"], [5], A.my_field.set(5 + A.my_field))
    _assert_expression(
        "SET #0.#1 = #2.#3 + :0",
        ["my_dict", "my_field", "my_dict", "my_field"],
        [5],
        A("my_dict.my_field").set(A("my_dict.my_field") + 5),
    )
    _assert_expression(
        "SET #0.#1 = :0 + #2.#3",
        ["my_dict", "my_field", "my_dict", "my_field"],
        [5],
        A("my_dict.my_field").set(5 + A("my_dict.my_field")),
    )


def test_set_plus_attribute():
    _assert_expression(
        "SET #0 = #1 + #2",
        ["my_field", "first_arg", "second_arg"],
        [],
        A.my_field.set(A.first_arg + A.second_arg),
    )

    _assert_expression(
        "SET #0.#1 = #2.#3 + #4.#5",
        ["dict_field", "my_field", "first_dict", "first_arg", "second_dict", "second_arg"],
        [],
        A("dict_field.my_field").set(A("first_dict.first_arg") + A("second_dict.second_arg")),
    )


def test_set_minus_value():
    _assert_expression("SET #0 = #1 - :0", ["my_field", "my_field"], [5], A.my_field.set(A.my_field - 5))
    _assert_expression("SET #0 = :0 - #1", ["my_field", "my_field"], [5], A.my_field.set(5 - A.my_field))
    _assert_expression(
        "SET #0.#1 = #2.#3 - :0",
        ["my_dict", "my_field", "my_dict", "my_field"],
        [5],
        A("my_dict.my_field").set(A("my_dict.my_field") - 5),
    )
    _assert_expression(
        "SET #0.#1 = :0 - #2.#3",
        ["my_dict", "my_field", "my_dict", "my_field"],
        [5],
        A("my_dict.my_field").set(5 - A("my_dict.my_field")),
    )


def test_set_minus_attribute():
    _assert_expression(
        "SET #0 = #1 - #2", ["my_field", "first_arg", "second_arg"], [], A.my_field.set(A.first_arg - A.second_arg)
    )
    _assert_expression(
        "SET #0.#1 = #2.#3 - #4.#5",
        ["my_dict", "my_field", "my_dict", "my_field", "another_dict", "another_field"],
        [],
        A("my_dict.my_field").set(A("my_dict.my_field") - A("another_dict.another_field")),
    )


def test_remove_attribute():
    _assert_expression("REMOVE #0", ["my_field"], [], A.my_field.remove())
    _assert_expression(
        "REMOVE #0, #1, #2",
        ["my_field", "second_field", "third_field"],
        [],
        A.my_field.remove(),
        A.second_field.remove(),
        A.third_field.remove(),
    )


def test_remove_index():
    _assert_expression("REMOVE #0[0]", ["my_list"], [], A.my_list.remove(0))
    _assert_expression("REMOVE #0[1]", ["my_list"], [], A.my_list.remove(1))
    _assert_expression("REMOVE #0.#1[0]", ["my_dict", "my_list"], [], A("my_dict.my_list").remove(0))
    _assert_expression("REMOVE #0.#1[1]", ["my_dict", "my_list"], [], A("my_dict.my_list").remove(1))


def test_cannot_remove_non_int():
    with pytest.raises(ValueError, match="Dyntastic remove\\(\\) update must be given an int, found 'str'"):
        A.my_list.remove("asd")  # type: ignore


def test_cannot_remove_int_subclass():
    class CustomInt(int):
        def __str__(self):
            return "gibberish"

    with pytest.raises(ValueError, match="Dyntastic remove\\(\\) update must be given an int, found 'CustomInt'"):
        A.my_list.remove(CustomInt(0))


def test_remove_attribute_and_index():
    _assert_expression("REMOVE #0, #1[2]", ["my_field", "my_list"], [], A.my_field.remove(), A.my_list.remove(2))


def test_add_int_value():
    _assert_expression("ADD #0 :0", ["my_int"], [5], A.my_int.add(5))
    _assert_expression("ADD #0 :0", ["my_int"], [-5], A.my_int.add(-5))


def test_add_decimal_value():
    value = Decimal("1.2")
    _assert_expression("ADD #0 :0", ["my_decimal"], [value], A.my_decimal.add(value))
    _assert_expression("ADD #0 :0", ["my_decimal"], [-value], A.my_decimal.add(-value))


def test_add_attribute():
    _assert_expression("ADD #0 #1", ["field", "other_field"], [], A.field.add(A.other_field))
    _assert_expression(
        "ADD #0 #1, #2 :0",
        ["field", "other_field", "third_field"],
        [10],
        A.field.add(A.other_field),
        A.third_field.add(10),
    )


def test_add_single_set_element():
    expected = ("ADD #0 :0", ["my_set"], [{"a"}])
    _assert_expression(*expected, A.my_set.add("a"))
    _assert_expression(*expected, A.my_set.add(["a"]))
    _assert_expression(*expected, A.my_set.add(("a")))


def test_add_multiple_set_elements():
    expected = ("ADD #0 :0", ["my_set"], [{1, 2}])
    _assert_expression(*expected, A.my_set.add({1, 2}))
    _assert_expression(*expected, A.my_set.add([1, 2]))
    _assert_expression(*expected, A.my_set.add((1, 2)))


def test_delete_single_set_element():
    expected = ("DELETE #0 :0", ["my_set"], [{"a"}])
    _assert_expression(*expected, A.my_set.delete("a"))
    _assert_expression(*expected, A.my_set.delete(["a"]))
    _assert_expression(*expected, A.my_set.delete(("a")))


def test_delete_multiple_set_elements():
    expected = ("DELETE #0 :0", ["my_set"], [{1, 2}])
    _assert_expression(*expected, A.my_set.delete({1, 2}))
    _assert_expression(*expected, A.my_set.delete([1, 2]))
    _assert_expression(*expected, A.my_set.delete((1, 2)))


def test_delete_attribute():
    _assert_expression("DELETE #0 #1", ["my_set", "my_other_set"], [], A.my_set.delete(A.my_other_set))



# pragma: nocover
from typing import TYPE_CHECKING, Any, Dict, Tuple, Type, TypeVar

import pydantic

_pydantic_major_version = int(pydantic.VERSION.split(".")[0])
IS_VERSION_1 = _pydantic_major_version == 1


BaseModelT = TypeVar("BaseModelT", bound=pydantic.BaseModel)

if TYPE_CHECKING:
    try:
        from pydantic.fields import ModelField as FieldInfo  # type: ignore[unused-ignore, attr-defined]
    except ImportError:
        from pydantic.fields import FieldInfo  # type: ignore[unused-ignore, assignment]

    def model_fields(model: Type[pydantic.BaseModel]) -> Dict[str, FieldInfo]:
        ...

    def model_dump(instance: pydantic.BaseModel, **kwargs) -> Dict[str, Any]:
        ...

    def annotation(field: FieldInfo) -> Any:
        ...

    def alias(field_name, field: Any) -> str:
        ...

    def to_jsonable_python(value: Any) -> Any:
        ...

    def try_model_construct(model: Type[BaseModelT], item: dict) -> Tuple[BaseModelT, bool]:
        ...

    class BaseModel(pydantic.BaseModel):
        ...

elif IS_VERSION_1:
    FieldInfo = pydantic.fields.ModelField

    def model_fields(model: Type[pydantic.BaseModel]) -> Dict[str, pydantic.fields.ModelField]:
        return model.__fields__

    def model_dump(instance: pydantic.BaseModel, **kwargs) -> Dict[str, Any]:
        return instance.dict(**kwargs)

    def annotation(field: FieldInfo) -> Any:
        return field.type_

    def alias(field_name, field: FieldInfo) -> str:
        return field.alias

    def to_jsonable_python(value: Any) -> Any:
        import json

        from pydantic.json import pydantic_encoder

        return json.loads(json.dumps(value, default=pydantic_encoder))

    def try_model_construct(model: Type[BaseModelT], item: dict) -> Tuple[BaseModelT, bool]:
        validated, fields_set, errors = pydantic.validate_model(model, item)
        if errors:
            # assume KEYS_ONLY or INCLUDE index
            fields_in_dynamo = {key: value for key, value in validated.items() if key in fields_set}
            data = model.construct(**fields_in_dynamo)
            return data, True
        else:
            data = model(**item)
            return data, False

    class BaseModel(pydantic.BaseModel):
        class Config:
            allow_population_by_field_name = True

else:
    from typing import Union

    try:
        # Python >= 3.8
        from typing import get_args, get_origin
    except ImportError:
        # Python 3.7
        def get_args(t):
            return getattr(t, "__args__", ())

        def get_origin(t):
            return getattr(t, "__origin__", None)

    FieldInfo = pydantic.fields.FieldInfo

    def model_fields(model: Type[pydantic.BaseModel]) -> Dict[str, pydantic.fields.FieldInfo]:
        return model.model_fields

    def model_dump(instance: pydantic.BaseModel, **kwargs) -> Dict[str, Any]:
        return instance.model_dump(**kwargs)

    def annotation(field: FieldInfo) -> Any:
        # Get rid of Optional[...] type if present (this is only used for
        # creating tables, where we want to know the actual inner type)
        type_ = field.annotation
        if get_origin(type_) is Union:
            args = get_args(type_)
            if len(args) == 2 and args[1] is type(None):  # noqa: E721
                return args[0]

        return type_

    def alias(field_name: str, field: FieldInfo) -> str:
        return field.alias or field_name

    def to_jsonable_python(value: Any) -> Any:
        import pydantic_core

        return pydantic_core.to_jsonable_python(value)

    class _FieldCollector(dict):
        pass

    def try_model_construct(model: Type[BaseModelT], item: dict) -> Tuple[BaseModelT, bool]:
        # Note: Hopefully there will be a better way to do this in the future
        # Related issue https://github.com/pydantic/pydantic/issues/7586

        collector = _FieldCollector()
        try:
            data = model.model_validate(item, context=collector)
            return data, False
        except pydantic.ValidationError:
            return model.model_construct(**collector), True

    class BaseModel(pydantic.BaseModel):
        model_config = pydantic.ConfigDict(populate_by_name=True)

        # TODO: eliminate the case where nested Dyntastic models overwrite the
        # top level _FieldCollector, potential to cause annoying bugs
        @pydantic.field_validator("*", mode="after")
        def collect_valid_fields(cls, v, info):
            if isinstance(info.context, _FieldCollector):
                info.context[info.field_name] = v
            return v


def field_type(model: Type[pydantic.BaseModel], field: str) -> Type:
    return annotation(model_fields(model)[field])


__all__ = [
    "BaseModel",
    "model_fields",
    "model_dump",
    "annotation",
    "alias",
    "to_jsonable_python",
    "try_model_construct",
    "field_type",
]



import pytest

from dyntastic.exceptions import DoesNotExist


def test_delete(item):
    item.delete()
    assert item.safe_get(item.id, getattr(item, "timestamp", None)) is None

    with pytest.raises(DoesNotExist):
        item.refresh()



import re
from typing import Optional, Type

import botocore.exceptions
import pytest

from dyntastic import A, Dyntastic, transaction
from dyntastic.transact import TRANSACTION_MAX_ITEMS, _transaction_writer_var


class _Table(Dyntastic):
    __table_name__ = "table"
    __hash_key__ = "hash_key"

    hash_key: str
    data: Optional[str] = None


@pytest.fixture
def Table():
    _Table.create_table(wait=True)
    return _Table


class _Table2(Dyntastic):
    __table_name__ = "table2"
    __hash_key__ = "hash_key"

    hash_key: str
    data: Optional[str] = None


@pytest.fixture
def Table2():
    _Table2.create_table(wait=True)
    return _Table2


@pytest.mark.parametrize("auto_commit", [True, False])
def test_no_items_exits_cleanly(auto_commit: bool):
    with transaction(auto_commit=auto_commit):
        assert isinstance(_transaction_writer_var.get(), transaction)

    assert _transaction_writer_var.get() is None


def test_multiple_regions_fails(Table: Type[_Table]):
    class Table2(Dyntastic):
        __table_name__ = "table2"
        __table_region__ = "us-west-2"
        __hash_key__ = "hash_key"

        hash_key: str

    Table2.create_table(wait=True)

    with transaction():
        Table(hash_key="foo").save()
        with pytest.raises(
            AssertionError,
            match=re.escape("All tables in a transaction must be in the same region (found None and us-west-2)"),
        ):
            Table2(hash_key="foo").save()

    assert _transaction_writer_var.get() is None


def test_item_limit_exceeded_without_auto_commit(Table: Type[_Table]):
    with pytest.raises(Exception, match=re.escape(f"Exceeded max items ({TRANSACTION_MAX_ITEMS}) in transaction")):
        with transaction():
            for i in range(TRANSACTION_MAX_ITEMS):
                Table(hash_key=f"foo{i}").save()

            Table(hash_key=f"foo{TRANSACTION_MAX_ITEMS + 1}").save()

    assert _transaction_writer_var.get() is None

    # No items should have been written since exception was raised
    assert not list(Table.scan())


def test_item_default_limit_exceeded_with_auto_commit(Table: Type[_Table]):
    with transaction(auto_commit=True) as w:
        for i in range(TRANSACTION_MAX_ITEMS):
            Table(hash_key=f"foo{i}").save()

        assert len(list(Table.scan())) == 100
        assert w.batches_submitted == 1
        assert w.items == []

        Table(hash_key=f"foo{TRANSACTION_MAX_ITEMS + 1}").save()

    assert _transaction_writer_var.get() is None

    assert len(list(Table.scan())) == 101


def test_item_custom_limit_exceeded_with_auto_commit(Table: Type[_Table]):
    with transaction(auto_commit=True, commit_every=50) as w:
        for i in range(50):
            Table(hash_key=f"foo{i}").save()

        assert len(list(Table.scan())) == 50
        assert w.batches_submitted == 1
        assert w.items == []

        Table(hash_key=f"foo{51}").save()

    assert _transaction_writer_var.get() is None

    assert len(list(Table.scan())) == 51


def test_commit_every_exceeds_max_items(Table: Type[_Table]):
    with pytest.raises(
        Exception,
        match=re.escape(f"commit_every cannot exceed DynamoDB limit of {TRANSACTION_MAX_ITEMS} items"),
    ):
        with transaction(auto_commit=True, commit_every=101):
            pass

    assert _transaction_writer_var.get() is None


def test_multiple_operations_on_same_item_errors(Table: Type[_Table]):
    # TODO: it seems moto deviates from AWS: moto allows a save and
    # update on the same key, but AWS does not

    with pytest.raises(
        botocore.exceptions.ClientError,
        match=re.escape("Transaction request cannot include multiple operations on one item"),
    ):
        with transaction():
            item = Table(hash_key="foo")
            item.save()
            item.save()

    assert _transaction_writer_var.get() is None

    assert not list(Table.scan())


def test_manual_commit(Table: Type[_Table]):
    with transaction() as w:
        item = Table(hash_key="foo")
        item.save()

        w.commit()
        assert w.batches_submitted == 1
        assert w.items == []

        items = list(Table.scan())
        assert len(items) == 1
        assert items[0].hash_key == "foo"
        assert items[0].data is None

        item.data = "some_data"
        item.save()

    assert _transaction_writer_var.get() is None

    items = list(Table.scan())
    assert len(items) == 1
    assert items[0].data == "some_data"
    assert w.batches_submitted == 2
    assert w.items == []


def test_batch_writer_errors_during_transaction(Table: Type[_Table]):
    with transaction():
        item = Table(hash_key="foo")
        item.save()

        with pytest.raises(ValueError, match=re.escape("Cannot use batch_writer() and transaction() at the same time")):
            with Table.batch_writer():
                item.update(A.data.set("bar"), refresh=False)

    assert _transaction_writer_var.get() is None

    items = list(Table.scan())
    assert len(items) == 1
    assert items[0].hash_key == "foo"
    assert items[0].data is None


def test_transaction_condition_errors_outside_of_transaction(Table: Type[_Table]):
    with pytest.raises(
        Exception, match=re.escape("Cannot use _Table.transaction_condition() outside of a transaction")
    ):
        Table(hash_key="foo").transaction_condition(A.hash_key.exists())


def test_update_warns_on_refresh(Table: Type[_Table]):
    item = Table(hash_key="foo")
    item.save()

    with transaction():
        with pytest.warns(UserWarning, match=re.escape("Cannot refresh model in transaction, skipping refresh")):
            item.update(A.data.set("bar"), refresh=True)

    assert _transaction_writer_var.get() is None

    with pytest.raises(
        ValueError,
        match=re.escape(
            "Dyntastic instance was not refreshed after update. "
            "Call refresh(), or use ignore_unrefreshed() to ignore safety checks"
        ),
    ):
        item.data

    items = list(Table.scan())
    assert len(items) == 1
    assert items[0].hash_key == "foo"
    assert items[0].data == "bar"


@pytest.mark.parametrize("condition", [None, A.hash_key.not_exists()])
@pytest.mark.parametrize("condition_check", [None, A.hash_key.exists()])
def test_save_single_table(Table: Type[_Table], condition, condition_check):
    with transaction() as w:
        item = Table(hash_key="foo")
        item.save(condition=condition)
        if condition_check:
            item.transaction_condition(condition_check)

    assert _transaction_writer_var.get() is None

    assert len(list(Table.scan())) == 1
    assert w.batches_submitted == 1
    assert w.items == []


@pytest.mark.parametrize("condition", [None, A.hash_key.not_exists()])
@pytest.mark.parametrize("condition_check", [None, A.hash_key.exists()])
def test_save_multiple_tables(Table: Type[_Table], Table2: Type[_Table2], condition, condition_check):
    with transaction() as w:
        item1 = Table(hash_key="foo")
        item1.save(condition=condition)
        item2 = Table2(hash_key="bar")
        item2.save(condition=condition)

        if condition_check:
            item1.transaction_condition(condition_check)
            item2.transaction_condition(condition_check)

    assert _transaction_writer_var.get() is None

    assert len(list(Table.scan())) == 1
    assert len(list(Table2.scan())) == 1
    assert w.batches_submitted == 1
    assert w.items == []


TRANSACTION_CANCELED_BASE_ERROR = (
    "An error occurred (TransactionCanceledException) when calling the "
    "TransactWriteItems operation: Transaction cancelled, please refer "
    "cancellation reasons for specific reasons"
)
SINGLE_CONDITION_FAILED_ERROR = re.escape(TRANSACTION_CANCELED_BASE_ERROR + " [ConditionalCheckFailed]")


def test_save_single_table_failing_condition(Table: Type[_Table]):
    with pytest.raises(
        botocore.exceptions.ClientError,
        match=SINGLE_CONDITION_FAILED_ERROR,
    ):
        with transaction():
            Table(hash_key="foo").save(condition=A.hash_key.exists())

    assert _transaction_writer_var.get() is None
    assert not list(Table.scan())


def test_save_single_table_failing_condition_check(Table: Type[_Table]):
    with pytest.raises(
        botocore.exceptions.ClientError,
        match=re.escape(TRANSACTION_CANCELED_BASE_ERROR + " [None, ConditionalCheckFailed]"),
    ):
        with transaction():
            item = Table(hash_key="foo")
            item.save()
            item2 = Table(hash_key="foo2")
            item2.transaction_condition(A.hash_key.exists())

    assert _transaction_writer_var.get() is None
    assert not list(Table.scan())


def test_save_multiple_tables_failing_condition(Table: Type[_Table], Table2: Type[_Table]):
    with pytest.raises(
        botocore.exceptions.ClientError,
        match=re.escape(TRANSACTION_CANCELED_BASE_ERROR + " [ConditionalCheckFailed, ConditionalCheckFailed]"),
    ):
        with transaction():
            Table(hash_key="foo").save(condition=A.hash_key.exists())
            Table2(hash_key="bar").save(condition=A.hash_key.exists())

    assert _transaction_writer_var.get() is None
    assert not list(Table.scan())


def test_save_multiple_tables_failing_condition_check(Table: Type[_Table], Table2: Type[_Table]):
    with pytest.raises(
        botocore.exceptions.ClientError,
        match=re.escape(TRANSACTION_CANCELED_BASE_ERROR + " [None, None, ConditionalCheckFailed]"),
    ):
        with transaction():
            Table(hash_key="foo").save()
            Table2(hash_key="foo2").save()
            item3 = Table(hash_key="bar")
            item3.transaction_condition(A.hash_key.exists())

    assert _transaction_writer_var.get() is None
    assert not list(Table.scan())


@pytest.mark.parametrize("condition", [None, A.hash_key.exists()])
@pytest.mark.parametrize("condition_check", [None, A.hash_key.not_exists()])
def test_update_single_table(Table: Type[_Table], condition, condition_check):
    Table(hash_key="foo").save()

    with transaction():
        item = Table.get("foo")
        item.update(A.data.set("bar"), condition=condition, refresh=False)
        if condition_check:
            Table(hash_key="foo2").transaction_condition(condition_check)

    assert _transaction_writer_var.get() is None

    items = list(Table.scan())
    assert len(items) == 1
    assert items[0].hash_key == "foo"
    assert items[0].data == "bar"


def test_update_single_table_failing_condition(Table: Type[_Table]):
    Table(hash_key="foo").save()

    with pytest.raises(
        botocore.exceptions.ClientError,
        match=SINGLE_CONDITION_FAILED_ERROR,
    ):
        with transaction():
            item = Table.get("foo")
            item.update(A.data.set("bar"), condition=A.hash_key.not_exists(), refresh=False)

    assert _transaction_writer_var.get() is None

    items = list(Table.scan())
    assert len(items) == 1
    assert items[0].hash_key == "foo"
    assert items[0].data is None


def test_update_single_table_failing_condition_check(Table: Type[_Table]):
    Table(hash_key="foo").save()

    with pytest.raises(
        botocore.exceptions.ClientError,
        match=re.escape(TRANSACTION_CANCELED_BASE_ERROR + " [None, ConditionalCheckFailed]"),
    ):
        with transaction():
            item = Table.get("foo")
            item.update(A.data.set("bar"), refresh=False)
            Table(hash_key="foo2").transaction_condition(A.hash_key.exists())

    assert _transaction_writer_var.get() is None

    items = list(Table.scan())
    assert len(items) == 1
    assert items[0].hash_key == "foo"
    assert items[0].data is None


@pytest.mark.parametrize("condition", [None, A.hash_key.exists()])
@pytest.mark.parametrize("condition_check", [None, A.hash_key.not_exists()])
def test_update_multiple_tables(Table: Type[_Table], Table2: Type[_Table2], condition, condition_check):
    Table(hash_key="foo").save()
    Table2(hash_key="bar").save()

    with transaction():
        item1 = Table.get("foo")
        item2 = Table2.get("bar")
        item1.update(A.data.set("data1"), condition=condition, refresh=False)
        item2.update(A.data.set("data2"), condition=condition, refresh=False)
        if condition_check:
            Table(hash_key="foo2").transaction_condition(condition_check)

    assert _transaction_writer_var.get() is None

    items = list(Table.scan())
    assert len(items) == 1
    assert items[0].hash_key == "foo"
    assert items[0].data == "data1"

    items2 = list(Table2.scan())
    assert len(items2) == 1
    assert items2[0].hash_key == "bar"
    assert items2[0].data == "data2"


def test_update_multiple_tables_failing_condition(Table: Type[_Table], Table2: Type[_Table2]):
    Table(hash_key="foo").save()
    Table2(hash_key="bar").save()

    with pytest.raises(
        botocore.exceptions.ClientError,
        match=re.escape(TRANSACTION_CANCELED_BASE_ERROR + " [ConditionalCheckFailed, ConditionalCheckFailed]"),
    ):
        with transaction():
            item1 = Table.get("foo")
            item2 = Table2.get("bar")
            item1.update(A.data.set("bar"), condition=A.hash_key.not_exists(), refresh=False)
            item2.update(A.data.set("bar"), condition=A.hash_key.not_exists(), refresh=False)

    assert _transaction_writer_var.get() is None

    items = list(Table.scan())
    assert len(items) == 1
    assert items[0].hash_key == "foo"
    assert items[0].data is None

    items2 = list(Table2.scan())
    assert len(items2) == 1
    assert items2[0].hash_key == "bar"
    assert items2[0].data is None


def test_update_multiple_tables_failing_condition_check(Table: Type[_Table], Table2: Type[_Table2]):
    Table(hash_key="foo").save()
    Table2(hash_key="bar").save()

    with pytest.raises(
        botocore.exceptions.ClientError,
        match=re.escape(TRANSACTION_CANCELED_BASE_ERROR + " [None, None, ConditionalCheckFailed]"),
    ):
        with transaction():
            item1 = Table.get("foo")
            item2 = Table2.get("bar")
            item1.update(A.data.set("bar"), refresh=False)
            item2.update(A.data.set("bar2"), refresh=False)
            Table(hash_key="foo2").transaction_condition(A.hash_key.exists())

    assert _transaction_writer_var.get() is None

    items = list(Table.scan())
    assert len(items) == 1
    assert items[0].hash_key == "foo"
    assert items[0].data is None

    items2 = list(Table2.scan())
    assert len(items2) == 1
    assert items2[0].hash_key == "bar"
    assert items2[0].data is None


@pytest.mark.parametrize("condition", [None, A.hash_key.exists()])
@pytest.mark.parametrize("condition_check", [None, A.hash_key.not_exists()])
def test_delete_single_table(Table: Type[_Table], condition, condition_check):
    Table(hash_key="foo").save()

    with transaction():
        item = Table.get("foo")
        item.delete(condition=condition)
        if condition_check:
            Table(hash_key="foo2").transaction_condition(condition_check)

    assert _transaction_writer_var.get() is None

    assert not list(Table.scan())


def test_delete_single_table_failing_condition(Table: Type[_Table]):
    Table(hash_key="foo").save()

    with pytest.raises(
        botocore.exceptions.ClientError,
        match=SINGLE_CONDITION_FAILED_ERROR,
    ):
        with transaction():
            item = Table.get("foo")
            item.delete(condition=A.hash_key.not_exists())

    assert _transaction_writer_var.get() is None

    items = list(Table.scan())
    assert len(items) == 1
    assert items[0].hash_key == "foo"
    assert items[0].data is None


def test_delete_single_table_failing_condition_check(Table: Type[_Table]):
    Table(hash_key="foo").save()

    with pytest.raises(
        botocore.exceptions.ClientError,
        match=re.escape(TRANSACTION_CANCELED_BASE_ERROR + " [None, ConditionalCheckFailed]"),
    ):
        with transaction():
            item = Table.get("foo")
            item.delete()
            Table(hash_key="foo2").transaction_condition(A.hash_key.exists())

    assert _transaction_writer_var.get() is None

    items = list(Table.scan())
    assert len(items) == 1
    assert items[0].hash_key == "foo"
    assert items[0].data is None


@pytest.mark.parametrize("condition", [None, A.hash_key.exists()])
@pytest.mark.parametrize("condition_check", [None, A.hash_key.not_exists()])
def test_delete_multiple_tables(Table: Type[_Table], Table2: Type[_Table2], condition, condition_check):
    Table(hash_key="foo").save()
    Table2(hash_key="bar").save()

    with transaction():
        item1 = Table.get("foo")
        item2 = Table2.get("bar")
        item1.delete(condition=condition)
        item2.delete(condition=condition)
        if condition_check:
            Table(hash_key="foo2").transaction_condition(condition_check)

    assert _transaction_writer_var.get() is None

    assert not list(Table.scan())
    assert not list(Table2.scan())


def test_delete_multiple_tables_failing_condition(Table: Type[_Table], Table2: Type[_Table2]):
    item = Table(hash_key="foo")
    item.save()
    item2 = Table2(hash_key="bar")
    item2.save()

    with pytest.raises(
        botocore.exceptions.ClientError,
        match=re.escape(TRANSACTION_CANCELED_BASE_ERROR + " [ConditionalCheckFailed, ConditionalCheckFailed]"),
    ):
        with transaction():
            item.delete(condition=A.hash_key.not_exists())
            item2.delete(condition=A.hash_key.not_exists())

    assert _transaction_writer_var.get() is None

    items = list(Table.scan())
    assert len(items) == 1
    assert items[0].hash_key == "foo"
    assert items[0].data is None

    items2 = list(Table2.scan())
    assert len(items2) == 1
    assert items2[0].hash_key == "bar"
    assert items2[0].data is None


def test_delete_multiple_tables_failing_condition_check(Table: Type[_Table], Table2: Type[_Table2]):
    item = Table(hash_key="foo")
    item.save()
    item2 = Table2(hash_key="bar")
    item2.save()

    with pytest.raises(
        botocore.exceptions.ClientError,
        match=re.escape(TRANSACTION_CANCELED_BASE_ERROR + " [None, None, ConditionalCheckFailed]"),
    ):
        with transaction():
            item.delete()
            item2.delete()
            Table(hash_key="foo2").transaction_condition(A.hash_key.exists())

    assert _transaction_writer_var.get() is None

    items = list(Table.scan())
    assert len(items) == 1
    assert items[0].hash_key == "foo"
    assert items[0].data is None

    items2 = list(Table2.scan())
    assert len(items2) == 1
    assert items2[0].hash_key == "bar"
    assert items2[0].data is None



import re
from datetime import datetime
from decimal import Decimal

import botocore.exceptions
import pytest

from dyntastic import A, Attr

from .conftest import MyNestedModel


def test_must_provide_actions(item):
    with pytest.raises(ValueError, match="Must provide at least one action to update"):
        item.update()


def test_set_new_attribute(item_no_my_str):
    item_no_my_str.update(A.my_str.set("bar"))
    assert item_no_my_str.my_str == "bar"


def test_set_existing_attribute(item):
    new_data = {
        "my_str": "new_value",
        "my_datetime": datetime(2022, 2, 13, 0, 0, 0),
        "my_bytes": b"new_bytes",
        "my_int": 6,
        "my_decimal": Decimal("2.5"),
        "my_str_set": {"d", "e"},
        "my_bytes_set": {b"e", b"f"},
        "my_int_set": {3, 4},
        "my_str_list": ["e", "f"],
        "my_int_list": [40, 50],
        "my_dict": {"d": 4, "e": 5},
        "my_nested_data": {"foo": ["bar", "baz", "bat"]},
        "my_nested_model": MyNestedModel(sample_field="updated"),
    }

    updates = [getattr(A, key).set(value) for key, value in new_data.items()]
    item.update(*updates)
    item.refresh()
    refreshed_data = dict(item)
    refreshed_data.pop("id")
    refreshed_data.pop("timestamp", None)
    assert refreshed_data == new_data


def test_no_refresh(item):
    item.update(A.my_str.set("bar"), refresh=False)
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Dyntastic instance was not refreshed after update. "
            "Call refresh(), or use ignore_unrefreshed() to ignore safety checks"
        ),
    ):
        item.my_str

    # make sure these attributes always are accessible
    assert item.ConditionException
    assert item.refresh
    assert item.ignore_unrefreshed

    item.refresh()
    assert item.my_str == "bar"


def test_ignore_refresh(item):
    item.update(A.my_str.set("bar"), refresh=False)
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Dyntastic instance was not refreshed after update. "
            "Call refresh(), or use ignore_unrefreshed() to ignore safety checks"
        ),
    ):
        item.my_str

    # make sure these attributes always are accessible
    assert item.ConditionException
    assert item.refresh
    assert item.ignore_unrefreshed

    item.ignore_unrefreshed()
    assert item.my_str == "foo"

    item.refresh()
    assert item.my_str == "bar"


def test_update_with_Attr(item):
    item.update(Attr.my_str.set("bar"))
    assert item.my_str == "bar"


def test_update_nested_string(item):
    item.update(A("my_nested_model.sample_field").set("new_value"))
    assert item.my_nested_model.sample_field == "new_value"


def test_set_and_remove_multiple_attributes(item):
    item.update(A.my_str.set("new_value"), A.my_int.set(500), A.my_decimal.remove(), A.my_dict.remove())
    assert item.my_str == "new_value"
    assert item.my_int == 500
    assert item.my_decimal is None
    assert item.my_dict is None


def test_set_if_not_exist_exists(item):
    item.update(A.my_str.set(A.my_str.if_not_exists("bar")))
    assert item.my_str == "foo"


def test_set_if_not_exist_does_not_exists(item_no_my_str):
    item_no_my_str.update(A.my_str.set(A.my_str.if_not_exists("bar")))
    assert item_no_my_str.my_str == "bar"


def test_set_default_exists(item):
    item.update(A.my_str.set_default("bar"))
    assert item.my_str == "foo"


def test_set_default_does_not_exist(item_no_my_str):
    item_no_my_str.update(A.my_str.set_default("bar"))
    assert item_no_my_str.my_str == "bar"


def test_set_int_plus(item):
    item.update(A.my_int.set(A.my_int + 100))
    assert item.my_int == 105


def test_set_int_rplus(item):
    item.update(A.my_int.set(100 + A.my_int))
    assert item.my_int == 105


def test_set_int_minus(item):
    item.update(A.my_int.set(A.my_int - 100))
    assert item.my_int == -95


def test_set_int_rminus(item):
    item.update(A.my_int.set(100 - A.my_int))
    assert item.my_int == 95


def test_list_append(item):
    item.update(A.my_str_list.set(A.my_str_list.list_append("z")))
    assert item.my_str_list == ["a", "b", "c", "d", "z"]


def test_list_append_if_not_exists(item_no_my_str_list):
    item = item_no_my_str_list

    default_list = A.my_str_list.if_not_exists(["y"])
    # TODO: fix type hinting around this pattern
    appended_list = A.list_append(default_list, "z")

    item.update(A.my_str_list.set(appended_list))
    assert item.my_str_list == ["y", "z"]


def test_append(item):
    item.update(A.my_str_list.append("z"))
    assert item.my_str_list == ["a", "b", "c", "d", "z"]


def test_remove_attribute(item):
    item.update(A.my_str.remove())
    assert item.my_str is None
    item.update(A.my_str_list.remove())
    assert item.my_str_list is None


def test_remove_list_element(item):
    item.update(A.my_str_list.remove(0))
    assert item.my_str_list == ["b", "c", "d"]


def test_add_single_string_to_set(item):
    item.update(A.my_str_set.add("z"))
    assert item.my_str_set == {"a", "b", "c", "z"}


@pytest.mark.parametrize("collection_type", [list, tuple, set])
def test_add_multiple_strings_to_set(collection_type, item):
    item.update(A.my_str_set.add(collection_type(["y", "z"])))
    assert item.my_str_set == {"a", "b", "c", "y", "z"}


def test_add_single_string_to_new_set(item_no_my_str_set):
    item = item_no_my_str_set
    item.update(A.my_str_set.add("z"))
    assert item.my_str_set == {"z"}


@pytest.mark.parametrize("collection_type", [list, tuple, set])
def test_add_multiple_strings_to_new_set(collection_type, item_no_my_str_set):
    item = item_no_my_str_set
    item.update(A.my_str_set.add(collection_type(["y", "z"])))
    assert item.my_str_set == {"y", "z"}


def test_cannot_add_int_to_set(item):
    with pytest.raises(
        botocore.exceptions.ClientError, match="An operand in the update expression has an incorrect data type"
    ):
        item.update(A.my_int_set.add(5))


def test_add_int(item):
    item.update(A.my_int.add(100))
    assert item.my_int == 105


def test_add_new_int(item_no_my_int):
    item = item_no_my_int
    item.update(A.my_int.add(100))
    assert item.my_int == 100


def test_add_decimal(item):
    item.update(A.my_decimal.add(Decimal("3.25")))
    assert item.my_decimal == Decimal("4.75")


def test_add_new_decimal(item_no_my_decimal):
    item = item_no_my_decimal
    item.update(A.my_decimal.add(Decimal("100.5")))
    assert item.my_decimal == Decimal("100.5")


def test_delete_single_string_from_set(item):
    item.update(A.my_str_set.delete("a"))
    assert item.my_str_set == {"b", "c"}


@pytest.mark.parametrize("collection_type", [list, tuple, set])
def test_delete_multiple_strings(collection_type, item):
    item.update(A.my_str_set.delete(collection_type(["a", "b"])))
    assert item.my_str_set == {"c"}


def test_delete_all_strings(item):
    item.update(A.my_str_set.delete(item.my_str_set))
    assert item.my_str_set is None


@pytest.fixture(params=["success", "failure", "raise_failure"])
def condition_tester(request):
    action = A.my_str.set("new_value")

    def tester(success_case, failure_case, item):
        if request.param == "success":
            item.update(action, condition=success_case, require_condition=True)
            assert item.my_str == "new_value"
        elif request.param == "failure":
            item.update(action, condition=failure_case)
            assert item.my_str == "foo"
        else:
            with pytest.raises(item.ConditionException(), match="The conditional request failed"):
                item.update(action, condition=failure_case, require_condition=True)

            item.refresh()
            assert item.my_str == "foo"

    return tester


nested_str = A("my_nested_model.sample_field")


@pytest.mark.parametrize(
    "success,failure",
    [
        # equality (all data types to check serialization)
        (A.my_str == "foo", A.my_str == "non-matching"),
        (A.my_str.eq("foo"), A.my_str.eq("non-matching")),
        (A.my_datetime == datetime(2022, 2, 12, 12, 26, 35), A.my_datetime == datetime(2022, 3, 1)),
        (A.my_bytes == b"foobar", A.my_bytes == b"not this"),
        (A.my_int == 5, A.my_int == -1000),
        (A.my_decimal == Decimal("1.5"), A.my_decimal == Decimal("100.5")),
        (A.my_str_set == {"a", "b", "c"}, A.my_str_set == {"x", "y"}),
        (A.my_bytes_set == {b"a", b"b", b"c"}, A.my_bytes_set == {b"x", b"y"}),
        (A.my_int_set == {1, 2}, A.my_int_set == {5, 6}),
        (A.my_str_list == ["a", "b", "c", "d"], A.my_str_list == ["x", "y", "z"]),
        (A.my_int_list == [10, 20, 30], A.my_int_list == [500, 600]),
        (A.my_dict == {"a": 1, "b": 2, "c": 3}, A.my_dict == {"x": 100, "y": 200}),
        (
            A.my_nested_data == [{"a": [{"foo": "bar"}], "b": "test"}, "some_string"],
            A.my_nested_data == ["not", {"this": "data"}],
        ),
        (nested_str == "hello", nested_str == "non-matching"),
        # inequality
        (A.my_str != "non-matching", A.my_str != "foo"),
        (A.my_str.ne("non-matching"), A.my_str.ne("foo")),
        (A.my_int != 5000, A.my_int != 5),
        (nested_str != "non-matching", nested_str != "hello"),
        (nested_str.ne("non-matching"), nested_str.ne("hello")),
        # </<=/>/>=
        (A.my_int < 10, A.my_int < 0),
        (A.my_int <= 5, A.my_int <= 4),
        (A.my_str < "zzz", A.my_str < "abc"),
        (A.my_str_set < {"a", "b", "c", "d"}, A.my_str_set < {"z"}),
        (A.my_int > 0, A.my_int > 10),
        (A.my_int >= 5, A.my_int >= 6),
        (A.my_str > "fff", A.my_str > "zzz"),
        (nested_str > "he", nested_str > "hello"),
        (nested_str >= "hello", nested_str >= "hello2"),
        (nested_str < "hello2", nested_str < "hello"),
        (nested_str <= "hello", nested_str <= "he"),
        # string operators
        (A.my_str.begins_with("f"), A.my_str.begins_with("z")),
        (A.my_str.between("f", "food"), A.my_str.between("z", "zebra")),
        (A.my_str.is_in({"foo", "bar", "baz"}), A.my_str.is_in({"d", "e"})),
        (A.my_str.contains("fo"), A.my_str.contains("blah")),
        (A.my_str_set.contains("a"), A.my_str_set.contains(A.my_str)),
        (nested_str.begins_with("h"), nested_str.begins_with("z")),
        (nested_str.between("h", "hello2"), nested_str.between("h", "he")),
        (nested_str.is_in({"a", "hello", "b"}), nested_str.is_in({"a", "b"})),
        (nested_str.contains("hel"), nested_str.contains("foo")),
        (A.my_str_set.contains("a"), A.my_str_set.contains(A.my_str)),
        # attribute_type/size/existence functions
        (A.my_str.attribute_type("S"), A.my_str.attribute_type("N")),
        (A.my_str.size < 15, A.my_str.size > 20),
        (A.my_str.size == 3, ~(A.my_str.size == 3)),
        (A.my_str.exists(), A.my_str.not_exists()),
        (nested_str.attribute_type("S"), nested_str.attribute_type("N")),
        (nested_str.size < 15, nested_str.size > 20),
        (nested_str.size == 5, ~(nested_str.size == 5)),
        (nested_str.exists(), nested_str.not_exists()),
        # logical operators
        (~(A.my_str == "bar"), ~(A.my_str == "foo")),
        ((A.my_str == "foo") | (A.my_str == "bar"), (A.my_str == "foo") & (A.my_int == 10)),
    ],
)
def test_update_condition(success, failure, condition_tester, item):
    condition_tester(success, failure, item)
